/*
 * dos2unix: Covert dos's text files to unix's ones
 * copyleft: June 26 1993, Guru International Inc.
 * Portable: Unix System, tested with Linux GCC 2.4.5
 * version:  1.1
 *
 * History:
 *      1.0 Init version, Jun 26, 1993
 *          - accept wildcard
 *          - can eliminated char 13 and 26
 *          - lightning speed
 *          - write to the original file
 *          - retain files status and permissions
 *          - automatically skip unregular file such as directory name
 *          - can report file not found msg
 *          - if no command line param inputed, show usage
 *          - No warning even -Wall of GCC
 *      1.1 Jun 28, 1993
 *          - AI technique for skipping possibly binary file, by using
 *            the fact that if first 8kb data have non-readable
 *            char(0-31) but not HT(9), LF(10) and CR(13), it must
 *            be binary file.
 *          - AI technique for skipping possible UNIX text file, for
 *            faster processing, no need to work with it, because
 *            it no needs to convert, by using the fact that if
 *            first 256 bytes if it has no CR(13) it must not be
 *            a text dos's text file. (A line of DOS's text file
 *            always less than 256 chars per line)
 *          - Skip if not a regular file.
 *          - Fix problem, now it can report 'directory skipped'
 *            correctly.
 *          - Add command line '-f' to force dos2unix to convert,
 *            no matter this file is binary or unix-already file.
 *            '-f-' is default mode, skip both above type files.
 *          - Add comand line '-h' to show usage, same as no parameter
 *            inputed.
 */
 
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* dummy name is used for internal DOS2UNIX */
#define HT                    9
#define LF                   10
#define FF                   12
#define CR                   13
#define CTRLZ                26

#define DUMMY_NAME           "dos2unix.tmp"
#define D2U_SUCCESS          1
#define D2U_FILE_NOT_FOUND   2
#define D2U_DIRECTORY        3
#define D2U_NONDATA_FILE     4
#define D2U_UNIX_ALREADY     5
#define D2U_NOT_REGULAR_FILE 6
#define BUFFER_SIZE          8192

/* bufin is buffer for input files.
 * bufout is buffer for output files.
 * statrec is record of status and permission of a file.
 * force is force-to-convert mode, init with false.
 */
int force = 0;
char *bufin, *bufout;
struct stat statrec;

/* prototype */	
int dos2unix(char *filename);
void usage(void);

int main(int argc, char **argv)
{
    int i, st;
	
	printf("\nDOS2UNIX 1.1, copyleft(w) by Guru International Inc.\n\n");
		
    if (argc == 1 || !strcmp(argv[1], "-h") ) {
		usage();
		return 0;
	}

	bufin = (char *)malloc(BUFFER_SIZE);
	bufout = (char *)malloc(BUFFER_SIZE);

	if (bufin == NULL || bufout == NULL) {
		printf("Internal Error: can't alloc memory\n");
		return 1;
	}
	
	unlink(DUMMY_NAME);
		
    for (i=1; i<argc; i++) {
/* turn on force mode */
        if (!strcmp(argv[i], "-f")) {
            force = 1;
            continue;
        }

/* turn of force mode */
        if (!strcmp(argv[i], "-f-")) {
            force = 0;
            continue;
        }
		
        printf("%-20s processing", argv[i]);
		st = dos2unix(argv[i]);
		
		switch (st) {
		case D2U_FILE_NOT_FOUND:
            printf("\r%-20s not found  \n", argv[i]);
			break;
		case D2U_SUCCESS:
            printf("\r%-20s done       \n", argv[i]);
			break;
		case D2U_DIRECTORY:
			sprintf(bufin, "%s/", argv[i]);
            printf("\r%-20s directory skipped\n", bufin);
			break;
		case D2U_NONDATA_FILE:
            printf("\r%-20s binary file skipped\n", argv[i]);
			break;
		case D2U_UNIX_ALREADY:
            printf("\r%-20s already-unix file skipped\n", argv[i]);
			break; 
        case D2U_NOT_REGULAR_FILE:
            printf("\r%-20s not regular file skipped\n", argv[i]);
            break;
        default:
			printf("\ninternal error: halt\b\n");
			break;
		}
	}

	unlink(DUMMY_NAME);
	
	free(bufin);
	free(bufout);	
	return 0;
} 

/* restore the previous file, due to it is a binary or unix-already file */
#define recall() { close(fdin); close(fdout); unlink(filename); \
        rename(DUMMY_NAME, filename); }

int dos2unix(char *filename)
{
	int fdin, fdout, i, j, size, dos_file = 0, ch;
	
/* skip if it is a directory or a not regular file */
	stat(filename, &statrec);
	if (S_ISDIR(statrec.st_mode)) 
		return D2U_DIRECTORY;
    if (!S_ISREG(statrec.st_mode))
        return D2U_NOT_REGULAR_FILE;

/* rename the source file to dummy file */
	rename(filename, DUMMY_NAME);
		
/* convertion is from dummy file back to source name one */ 
 
	if ( (fdin = open(DUMMY_NAME, O_RDONLY)) < 0)
		return D2U_FILE_NOT_FOUND;
	fdout = creat(filename, statrec.st_mode);

/* read first chunk, to see if it is a binary or not regular or unix-already
 * skip it
 */
    size = read(fdin, bufin, BUFFER_SIZE);
	for (i=0, j=0; i<size; i++) {
		ch = bufin[i];
		if (ch != CR && ch != CTRLZ) {
			bufout[j++] = ch;
        } else if (i < size)    /* it is a dos file if found CR(13)        */
            dos_file = 1;       /* or CTRLZ(26) but must not the last char */

/* if it a unreadable char(text unrelated char), it is binary */
        if (ch != LF && ch != CR && ch != HT && ch != FF &&
				ch < 32  && !force) {
			recall();
#ifdef DEBUG
			printf("\nchar = %d\n", ch); 
#endif
			return D2U_NONDATA_FILE;
		}
/* if not found any CR(13) in first 256 bytes file, it must be unix-already */
		if (i == 256 && !dos_file && !force) {
			recall();
			return D2U_UNIX_ALREADY;
		}
	}
		
/* trap the file that < 256 bytes but it is unix-already file */
    if (!dos_file && !force) {
		recall();
		return D2U_UNIX_ALREADY;
	}
	
/* save first chunk */
    write(fdout, bufout, j);

/* read each input file a chunk to bufin, trimmed char CR(13), CTRLZ(26) out
 * to output file, loop until eof
 */ 
	while (size == BUFFER_SIZE) {
		size = read(fdin, bufin, BUFFER_SIZE);
		for (i=0, j=0; i<size; i++) {
			ch = bufin[i];
			if (ch != CR && ch != CTRLZ)
				bufout[j++] = ch;
		}
		write(fdout, bufout, j);
	} 
	close(fdin);
	close(fdout);
		
	return D2U_SUCCESS;
}
	

/* show usage, if no commandline parameter inputed */	
void usage(void)
{
	printf("dos2unix is a utility for converting DOS's" 
            "text files to UNIX ones.\n"
            "usage: dos2unix [-f] [-h] file [[-f] [-f-] file ...]\n"
            "       -f  is force mode, it is forced to convert every file.\n"
            "       -f- is normal mode, skip binary or a UNIX-already file.\n"
            "       -h  is for display this help.\n\n"
            "       Wildcard is allowed\n\n");
}


